#include <afblib/strhash.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "lockmanager.h"

struct queue_member {
   struct user* user;
   lockhandler handler;
   void* handle;
   struct queue_member* next;
};

struct lock {
   char* name;
   struct queue_member* head; /* first element is the lock holder */
   struct queue_member* tail;
};

struct user {
   char* name;
   strhash locks; /* all locks were user is either the holder or in the queue */
};

struct lockset {
   strhash users; /* value: struct user* */
   strhash locks; /* value: struct lock* */
};

typedef struct lockset* lockset[1];
typedef void (*lockhandler)(lockset set, char* lock_name, void* handle);

/* initialize lock set */
bool lm_init(lockset set) {
   struct lockset* ls = calloc(1, sizeof(struct lockset));
   if (ls) {
      if (strhash_alloc(&ls->users, 32)) {
	 if (strhash_alloc(&ls->locks, 32)) {
	    set[0] = ls;
	 } else {
	    strhash_free(&ls->users); free(ls); ls = 0;
	 }
      } else {
	 free(ls); ls = 0;
      }
   }
   return ls != 0;
}

/* release storage */
void lm_free(lockset set) {
   struct lockset* ls = set[0];
   if (!ls) return;

   strhash_start(&ls->users); char* user_name;
   while (strhash_next(&ls->users, &user_name)) {
      char* value;
      if (strhash_lookup(&ls->users, user_name, &value)) {
	 struct user* user = (struct user*) value;
	 if (user) {
	    free(user->name);
	    strhash_free(&user->locks);
	    free(user);
	 }
      }
   }

   strhash_start(&ls->locks); char* lock_name;
   while (strhash_next(&ls->locks, &lock_name)) {
      char* value;
      if (strhash_lookup(&ls->locks, lock_name, &value)) {
	 struct lock* lock = (struct lock*) lock;
	 if (lock) {
	    struct queue_member* member = lock->head;
	    while (member) {
	       struct queue_member* old = member;
	       member = member->next;
	       free(old);
	    }
	    free(lock->name);
	    free(lock);
	 }
      }
   }
   free(ls);
   set[0] = 0;
}

/* add user to the lock set */
bool lm_adduser(lockset set, char* user_name) {
   struct lockset* ls = set[0];
   if (!ls) return false;
   if (strhash_exists(&ls->users, user_name)) return false;
   struct user* user = calloc(1, sizeof(struct user));
   if (!user) return false;
   user->name = strdup(user_name);
   if (!user->name) {
      free(user); return false;
   }
   if (!strhash_alloc(&user->locks, 8)) {
      free(user->name); free(user); return false;
   }
   if (!strhash_add(&ls->users, user->name, (char*) user)) {
      strhash_free(&user->locks);
      free(user->name); free(user);
      return false;
   }
   return true;
}

/* remove user and all his/her locks */
bool lm_removeuser(lockset set, char* user_name) {
   struct lockset* ls = set[0];
   if (!ls) return false;
   char* value;
   if (!strhash_lookup(&ls->users, user_name, &value)) return false;
   struct user* user = (struct user*) value;
   if (user) {
      strhash_start(&user->locks);
      char* lock_name;
      while (strhash_next(&user->locks, &lock_name)) {
	 lm_release(set, user->name, lock_name);
      }
      free(user->name);
      strhash_free(&user->locks);
      free(user);
   }
   return strhash_remove(&ls->users, user_name);
}

bool lm_stat(lockset set, char* lock_name, stralloc* holder) {
   struct lockset* ls = set[0];
   if (!ls) return false;
   char* value;
   if (!strhash_lookup(&ls->locks, lock_name, &value)) {
      /* unknown locks are considered to be free */
      holder->len = 0; return true;
   }
   struct lock* lock = (struct lock*) value;
   if (lock->head == 0) {
      holder->len = 0; return true;
   }
   stralloc_copys(holder, lock->head->user->name);
   return true;
}

lock_status lm_lock(lockset set, char* user_name, char* lock_name,
	            lockhandler handler, void* handle) {
   struct lockset* ls = set[0];
   if (!ls) return LOCK_FAILED;

   char* value;
   struct user* user;
   if (!strhash_lookup(&ls->users, user_name, &value)) {
      return LOCK_FAILED;
   }
   user = (struct user*) value;

   /* already locked? */
   if (strhash_exists(&user->locks, lock_name)) {
      return LOCK_FAILED;
   }

   struct lock* lock;
   if (strhash_lookup(&ls->locks, lock_name, &value)) {
      lock = (struct lock*) value;
   } else {
      lock = calloc(1, sizeof(struct lock));
      if (!lock) return LOCK_FAILED;
      lock->name = strdup(lock_name);
      if (!strhash_add(&ls->locks, lock->name, (char*) lock)) {
	 free(lock->name); free(lock);
	 return LOCK_FAILED;
      }
   }

   if (!strhash_add(&user->locks, lock->name, 0)) {
      return LOCK_FAILED;
   }

   struct queue_member* member = calloc(1, sizeof(struct queue_member));
   if (!member) return LOCK_FAILED;
   member->user = user;
   member->handler = handler;
   member->handle = handle;
   if (lock->head) {
      lock->tail->next = member;
   } else {
      lock->head = member;
   }
   lock->tail = member;

   if (lock->head == member) {
      return LOCK_OK;
   } else {
      return LOCK_WAITING;
   }
}

bool lm_release(lockset set, char* user_name, char* lock_name) {
   struct lockset* ls = set[0];
   if (!ls) return false;

   char* value;
   struct user* user;
   if (!strhash_lookup(&ls->users, user_name, &value)) {
      return false;
   }
   user = (struct user*) value;
   if (!strhash_exists(&user->locks, lock_name)) {
      return false;
   }
   if (!strhash_remove(&user->locks, lock_name)) {
      return false;
   }

   struct lock* lock;
   if (!strhash_lookup(&ls->locks, lock_name, &value)) {
      return false;
   }
   lock = (struct lock*) value;
   if (!lock) return false;

   struct queue_member* member = lock->head;
   if (!member) return false;

   if (member->user == user) {
      /* release by lock holder */
      lock->head = member->next;
      if (!lock->head) lock->tail = 0;
      free(member);
      member = lock->head;
      if (member) {
	 /* notify next member in queue */
	 (*(member->handler))(set, lock->name, member->handle);
      }
   } else {
      /* release by someone who is in the queue but not holding the lock */
      struct queue_member* prev = 0;
      while (member && member->user != user) {
	 prev = member; member = member->next;
      }
      if (prev && member) {
	 prev->next = member->next;
	 free(member);
      }
      strhash_remove(&user->locks, lock_name);
   }
   return true;
}
