#include <assert.h>
#include <stdlib.h>
#include <stralloc.h>
#include <string.h>
#include "mxpsession.h"
#include "mxpresponse.h"
#include "mxprequest.h"
#include "lockmanager.h"

#define EQUAL(sa,str) (strncmp((sa.s),(str),(sa.len)) == 0)

struct mxp_session {
   enum {START, IDENTIFIED} state;
   stralloc name;
   mxp_request request;
   mxp_response response;
   lockset locks;
};

void mxp_session_open(session* s) {
   struct mxp_session* ms = calloc(1, sizeof(struct mxp_session));
   if (!ms) {
      close_session(s); return;
   }
   s->handle = ms;
   ms->state = START;
   ms->response.status = MXP_SUCCESS;
   ms->locks[0] = ((struct lockset**) s->global_handle)[0];
   if (!write_mxp_response(s, &ms->response)) {
      close_session(s); return;
   }
}

void mxp_session_lock(lockset set, char* lock_name, void* handle) {
   session* s = handle; assert(s);
   struct mxp_session* ms = s->handle; assert(ms);
   ms->response.status = MXP_SUCCESS;
   stralloc_copys(&ms->response.message, "locked");
   if (!write_mxp_response(s, &ms->response)) {
      close_session(s);
   }
}

void mxp_session_read(session* s) {
   struct mxp_session* ms = s->handle; assert(ms);
   if (!read_mxp_request(s, &ms->request)) {
      close_session(s); return;
   }
   if (ms->state == START) {
      if (EQUAL(ms->request.keyword, "id")) {
	 if (stralloc_0(&ms->request.parameter) &&
	       lm_adduser(ms->locks, ms->request.parameter.s)) {
	    ms->response.status = MXP_SUCCESS;
	    stralloc_copy(&ms->name, &ms->request.parameter);
	    ms->state = IDENTIFIED;
	    stralloc_copys(&ms->response.message, "welcome");
	 } else {
	    ms->response.status = MXP_FAILURE;
	    stralloc_copys(&ms->response.message, "identification failed");
	 }
      } else {
	 ms->response.status = MXP_FAILURE;
	 stralloc_copys(&ms->response.message, "identify yourself first");
      }
   } else if (EQUAL(ms->request.keyword, "lock")) {
      if (stralloc_0(&ms->request.parameter)) {
	 switch (lm_lock(ms->locks, ms->name.s, ms->request.parameter.s,
	       mxp_session_lock, s)) {
	    case LOCK_FAILED:
	       ms->response.status = MXP_FAILURE;
	       stralloc_copys(&ms->response.message, "unable to lock");
	       break;
	    case LOCK_OK:
	       ms->response.status = MXP_SUCCESS;
	       stralloc_copys(&ms->response.message, "locked");
	       break;
	    case LOCK_WAITING:
	       ms->response.status = MXP_CONTINUATION;
	       stralloc_copys(&ms->response.message, "waiting");
	       break;
	 }
      } else {
	 ms->response.status = MXP_FAILURE;
	 stralloc_copys(&ms->response.message, "out of memory");
      }
   } else if (EQUAL(ms->request.keyword, "release")) {
      if (stralloc_0(&ms->request.parameter) &&
	    lm_release(ms->locks, ms->name.s, ms->request.parameter.s)) {
	 ms->response.status = MXP_SUCCESS;
	 stralloc_copys(&ms->response.message, "released");
      } else {
	 ms->response.status = MXP_FAILURE;
	 stralloc_copys(&ms->response.message, "release failed");
      }
   } else if (EQUAL(ms->request.keyword, "stat")) {
      stralloc holder = {0};
      if (stralloc_0(&ms->request.parameter) &&
	    lm_stat(ms->locks, ms->request.parameter.s, &holder)) {
	 if (holder.len == 0) {
	    ms->response.status = MXP_SUCCESS;
	    stralloc_copys(&ms->response.message, "free");
	 } else {
	    ms->response.status = MXP_CONTINUATION;
	    stralloc_copy(&ms->response.message, &holder);
	    if (!write_mxp_response(s, &ms->response)) {
	       close_session(s); return;
	    }
	    ms->response.status = MXP_SUCCESS;
	    stralloc_copys(&ms->response.message, "held");
	 }
      } else {
	 ms->response.status = MXP_FAILURE;
	 stralloc_copys(&ms->response.message, "stat failed");
      }
   } else {
      ms->response.status = MXP_FAILURE;
      stralloc_copys(&ms->response.message, "unknown command");
   }
   if (!write_mxp_response(s, &ms->response)) {
      close_session(s);
   }
}

void mxp_session_hangup(session* s) {
   struct mxp_session* ms = s->handle;
   if (ms) {
      if (ms->state == IDENTIFIED) lm_removeuser(ms->locks, ms->name.s);
      stralloc_free(&ms->name);
      free_mxp_request(&ms->request);
      free_mxp_response(&ms->response);
   }
}
